import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tqdm import tqdm
from collections import defaultdict

from quapy.data import LabelledCollection
from quapy.protocol import UPP

# this script computes the distribution of smoothness/sharpness of the books;
# either considering all books, as well as considering different groups of books' reviews (by product)
# Mirko asked for some exploration of values (votes, num reviews), and percentiles of dataset shift as measured in terms
# of NMD between training set prevalences and sample prevalences; this script does this
# It also generates a csv containing all the prevalence values by product


read_meta = True


def not_smoothness(p):
    return 0.5 * sum((-p_prev + 2*p_i - p_next)**2 for p_prev, p_i, p_next in zip(p[:-2], p[1:-1], p[2:]))


def _check_arrays(prevs):
    prevs = np.asarray(prevs)
    if prevs.ndim==1:
        prevs = prevs.reshape(1,-1)
    return prevs


# mean normalized match distance
def mnmd(prevs, prevs_hat):
    prevs = _check_arrays(prevs)
    prevs_hat = _check_arrays(prevs_hat)
    assert prevs.shape == prevs_hat.shape, f'wrong shape; found {prevs.shape} and {prevs_hat.shape}'

    nmds = [nmd(p, p_hat) for p, p_hat in zip(prevs, prevs_hat)]
    return np.mean(nmds)


# normalized match distance
def nmd(prev, prev_hat):
    n = len(prev)
    return (1./(n-1))*mdpa(prev, prev_hat)


"""
Minimum Distance of Pair Assignments (MDPA) [cha2002measuring] for ordinal pdfs `a` and `b`.
The MDPA is a special case of the Earth Mover's Distance [rubner1998metric] that can be
computed efficiently.
[Mirko Bunse's code from Julia adapted]
"""
def mdpa(a, b):
    assert len(a) == len(b), "histograms have to have the same length"
    assert np.isclose(sum(a), sum(b)), "histograms have to have the same mass (difference is $(sum(a)-sum(b))"

    # algorithm 1 in [cha2002measuring]
    prefixsum = 0.0
    distance  = 0.0
    for i in range(len(a)):
        prefixsum += a[i] - b[i]
        distance += abs(prefixsum)

    return distance / sum(a)  # the normalization is a fix to the original MDPA


def prepare_vote_field(df):
    df['vote'] = df['vote'].fillna('0')
    df['vote'] = df['vote'].apply(lambda x: x.replace(',', ''))
    df['vote'] = pd.to_numeric(df['vote'])
    return df


def read_from_huge_json(filein):
    df = pd.read_json(filein, lines=True)
    df.drop(columns=[
        'verified', 'reviewTime', 'reviewerID', 'style', 'reviewerName', 'reviewText', 'summary', 'unixReviewTime',
        'image'
    ], inplace=True)

    df = prepare_vote_field(df)
    return df


def read_from_metadata(filein):
    df = pd.read_csv(filein)
    df['vote'] = pd.to_numeric(df['vote'])
    return df


def filter_by_vote(df, vote_threshold=1):
    df = df[df['vote'] >= vote_threshold]
    df.drop(columns=['vote'], inplace=True)
    return df


if read_meta:
    filein = '/media/moreo/Volume/Datasets/Amazon/meta/Books.csv'
    readfn = read_from_metadata
else:
    filein='/media/moreo/Volume/Datasets/Amazon/raw/Books.json'
    readfn = read_from_huge_json


def create_dictionary_bookid_ratings(df):
    # pass to dictionaries
    df = df.reset_index()  # make sure indexes pair with number of rows

    ids = df['asin'].values
    overalls = df['overall'].values

    # Defining a dict
    d = defaultdict(list)
    for i, id in tqdm(enumerate(ids), total=len(ids), desc='passing to dictionary'):
        d[id].append(overalls[i])

    return d

def get_stats(distribution, msg=''):
    # computes the mean, max, min, perc5, perc25, perc50, perc75, perc95 of the distribution
    vmean = np.mean(distribution)
    vmax = np.max(distribution)
    vmin = np.min(distribution)
    q05 = np.percentile(distribution, 5)
    q25 = np.percentile(distribution, 25)
    q50 = np.percentile(distribution, 50)
    q75 = np.percentile(distribution, 75)
    q95 = np.percentile(distribution, 95)
    print(f'{msg}: percentiles {q05:.5f}\t{q25:.5f}\t{q50:.5f}\t{q75:.5f}\t{q95:.5f}')
    print(f'{msg}: ave={np.mean(distribution):.5f}')
    print(f'{msg}: max={np.max(distribution):.5f}')
    print(f'{msg}: min={np.min(distribution):.5f}')
    return vmean, vmax, vmin, q05, q25, q50, q75, q95

with open('book_stats.csv', 'wt') as foo:
    foo.write(f'minvotes\tminreviews\t#products\t#reviews'
              f'\tsharp-ave\tsharp-max\tsharp-min\t'
              f'sharp-P5\tsharp-P25\tsharp-P50\tsharp-P75\tsharp-P95'
              f'\tshift-ave\tshift-max\tshift-min\t'
              f'shift-P5\tshift-P25\tshift-P50\tshift-P75\tshift-P95'
              f'\n')

    for votes_support in [1]:

        df = readfn(filein)
        df = df[df['overall']>0]  # there are a couple of reviews with 0 stars (the min should be 1)

        num_entries = len(df)
        df = filter_by_vote(df, vote_threshold=votes_support)
        num_entries_with_vote = len(df)

        unique_product_ids = df['asin'].unique()
        num_products = len(unique_product_ids)

        print(df.columns)
        print(f'num rows {len(df)} (before vote-thresholding {num_entries}, after thresholding {num_entries_with_vote})')
        print(f'num unique products {num_products}')

        d = create_dictionary_bookid_ratings(df)

        for reviews_support in [100]:
            with open(f'./prevalence_votes{votes_support}_reviews{reviews_support}.csv', 'wt') as fprev: 
                sharpness_all = []
                num_products_with_reviews = 0
                sel_ids, sel_overalls = [], []
                for product_id, ratings in tqdm(d.items(), total=len(d), desc='processing histograms'):
                    n_ratings = len(ratings)
                    if n_ratings >= reviews_support:
                        sel_ids.extend([product_id] * n_ratings)
                        sel_overalls.extend(ratings)

                        prev = np.histogram(ratings, bins=np.array([0, 1, 2, 3, 4, 5]) + 0.5, density=True)[0]
                        for i, prev_i in enumerate(prev):
                            fprev.write(f'{prev_i:.5f}')
                            if i < len(prev)-1:
                                fprev.write('\t')
                            else:
                                fprev.write('\n')
                        sharpness = not_smoothness(prev)
                        sharpness_all.append(sharpness)
                        num_products_with_reviews+=1

            print(f'#votes-support (min number of votes): {votes_support}')
            print(f'#reviews with >#votes-support: {num_entries_with_vote}/{num_entries}={100*num_entries_with_vote/num_entries:.2f}%')

            print(f'#reviews-support (min number of reviews): {reviews_support}')
            print(f'#products with >#reviews-support: {num_products_with_reviews}/{num_products}={100*num_products_with_reviews/num_products:.2f}%')

            vmean, vmax, vmin, q05, q25, q50, q75, q95 = get_stats(sharpness_all, 'sharpness')

            allbooks_prev = np.histogram(sel_overalls, bins=np.array([0, 1, 2, 3, 4, 5]) + 0.5, density=True)[0]
            allbooks_sharpness = not_smoothness(allbooks_prev)
            print(f'all books prev={allbooks_prev} has sharpness {allbooks_sharpness:.4f}')

            sel_collection = LabelledCollection(instances=sel_ids, labels=sel_overalls, classes=[1,2,3,4,5])
            prot = UPP(sel_collection, sample_size=1000, repeats=5000)
            prot_iterator = prot()
            shifts = []
            for _, test_prev in tqdm(prot_iterator, total=prot.total()):
                shifts.append(nmd(allbooks_prev, prev_hat=test_prev))
            s_mean, s_max, s_min, s_q05, s_q25, s_q50, s_q75, s_q95 = get_stats(shifts, 'shift')

            foo.write(f'{votes_support}\t{reviews_support}\t{num_products_with_reviews}\t{len(sel_ids)}'
                    f'\t{vmean:.5f}\t{vmax:.5f}\t{vmin:.5f}\t'
                    f'{q05:.5f}\t{q25:.5f}\t{q50:.5f}\t{q75:.5f}\t{q95:.5f}'
                    f'\t{s_mean:.5f}\t{s_max:.5f}\t{s_min:.5f}\t'
                    f'{s_q05:.5f}\t{s_q25:.5f}\t{s_q50:.5f}\t{s_q75:.5f}\t{s_q95:.5f}\n')


